/* $Id: class.cc,v 1.25 2002/03/17 19:09:36 martin Exp $ */
// Medoosa: Support for Documentation of C++ Programs
// Master's Thesis, Faculty of Mathematics and Physics, Charles University
// Martin Vidner, 2001
// Released under the terms of the GNU General Public License.
#define _ISOC99_SOURCE
// snprintf
#include <stdlib.h>
#include "class.h"
#include <libxml/xpath.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include "debug.h"

#include <algorithm>

void MGeneralization::LoadTokens(list<string>::iterator b, list<string>::iterator e) {
  list<string>::iterator i = b;
  _isVirtual = false;
  _visibilityToken.Read("private", false);

  // there may be "virtual" and "public"/"protected"/"private", in any order
  if (*i == "virtual") {
    _isVirtual = true;
    ++i;
    if (_visibilityToken.Read(i->c_str()))
      ++i;
  }
  else if (_visibilityToken.Read(i->c_str())) {
    ++i;
    if (*i == "virtual") {
      _isVirtual = true;
      ++i;
    }
  }

  list<string>::iterator templateInstanceParams = find(i, e, "<");
  _parent = Detokenize(list<string>(i, templateInstanceParams));

  if (mDebug.IsEnabled('g')) {
    fprintf(stderr, " Parent: %s\n", _parent.c_str());
    if (templateInstanceParams != e)
      fprintf(stderr, "  (Actually, %s)\n", 
	      Detokenize(list<string>(i, e)).c_str());
  }
}

void MGeneralization::LoadUxf(xmlNodePtr ugener) {
  _isVirtual = false;
  xmlNodePtr uchild;
  char *s;
  for (uchild = ugener->children; uchild != NULL; uchild = uchild->next) {
    if (uchild->type == XML_ELEMENT_NODE) {
      if (!xmlStrcmp(uchild->name, IS_UTF8 "Parent")) {
	s = CAST_BAD xmlNodeGetContent(uchild);
	_parent = s;
	if (mDebug.IsEnabled('g'))
	  fprintf(stderr, "Child: %s, Parent: %s\n", _child->GetName().c_str(), s);
	free(s);
      }
      else if (!xmlStrcmp(uchild->name, IS_UTF8 "TaggedValue")) {
	xmlNodePtr node;
	string tag;
	node = GetTaggedNode(uchild, &tag);
	if (tag == "DiagramRoute") {
	  _route.LoadDia(node);
	}
      }
    }
  }
}

xmlNodePtr MGeneralization::SaveUxf() const {
  xmlNodePtr ugeneralization;
  xmlNodePtr unode;
  ugeneralization = xmlNewNode(NULL, IS_UTF8 "Generalization");

  // (this is a UXF extension)
  xmlAddChild(ugeneralization, _route.SaveDia());
  
  // (this is a UXF extension)
  xmlAddChild(ugeneralization, 
	      NewTaggedValue(IS_UTF8 "Visibility",
			     BAD_CAST _visibilityToken.Write()));

  if (_isVirtual) {
    unode = xmlNewChild(ugeneralization, NULL, IS_UTF8 "Stereotype", NULL);
    xmlNewChild(unode, NULL, IS_UTF8 "Name", IS_UTF8 "virtual");
  }

  // get the parent's qualified name, if possible (it may be a library class)
  MModelElement *parent = _child->GetElement(_parent);
  string parentQn = parent? parent->GetQualifiedName() : _parent;

  xmlNewChild(ugeneralization, NULL, IS_UTF8 "Parent", 
	      BAD_CAST parentQn.c_str());  
  xmlNewChild(ugeneralization, NULL, IS_UTF8 "Child", 
	      BAD_CAST _child->GetQualifiedName().c_str());

  return ugeneralization;
}

// ---------------------------------------------------------------

MClass::MClass(MNamespace *ns): MNamespace("", ns), _isAbstract(false) {
}

MClass::MClass(const MClass& b): MNamespace(b) {
  cerr << "class copy\n";  //TODO
}

/*
MClass& MClass::operator = (const MClass&)
{
  //TODO
  return *this;
}
*/

static void Warning(const char *s) {
  fprintf(stderr,"Warning: %s\n", s);
}

string MClass::LoadName(xmlNodePtr cnode) {
  xmlChar *xname;
  string sname;

  xname = xmlGetProp(cnode, IS_UTF8 "name");
  sname = CAST_BAD xname;
  free(xname);
  return sname;
}

void MClass::LoadParseTree(xmlNodePtr cclass, const string &fileName) {
  _name = LoadName(cclass);
  if (mDebug.IsEnabled('c'))
    fprintf(stderr, "%s %s\n", cclass->name, _name.c_str());

  // find the header
  list<string> header;
  header = LoadTokens(FindFirstTokenChild(cclass));
  list<string>::iterator beg = header.begin();
  list<string>::iterator end = header.end();

  // source file and line
  _containingFiles.insert(fileName);
  _sourceInfo.LoadCtfx(cclass);

  // parse generalizations (superclasses)
  list<string>::iterator superClassesBegin;
  superClassesBegin = find(beg, end, ":");
  if (superClassesBegin != end) {
    ++superClassesBegin;
    // the same trick... should generalize it :)
    list<string>::iterator i = superClassesBegin;
    while (i != end && *i->c_str() != '{') {
      // set [superClassBegin,i) range 
      list<string>::iterator superClassBegin = i;
      char c;
      while (i != end, c = * i->c_str(), c != ',' && c != '{') {
	if (c == '(')
	  ForwardSkipBrackets(i, '(', ')', end);
	else if (c == '<')
	  ForwardSkipBrackets(i, '<', '>', end);
	else
	  ++i;
      }
      MGeneralization gen(this);
      gen.LoadTokens(superClassBegin, i);
      _superClasses.push_back(gen);
      // skip ','
      if (* i->c_str() != '{')
	++i;
    }
  }
  
  // parse template parameters
  if(!xmlStrcmp(cclass->name, IS_UTF8 "tmc")) {
    list<string>::iterator paramsBegin = find(beg, end, "<");
    list<string>::iterator paramsEnd = paramsBegin;
    ForwardSkipBrackets(paramsEnd, '<', '>', end);

    // Common part for function and template parameters, '>', ')'
    list<string>::iterator i = paramsBegin;
    ++paramsBegin; // into the brackets
    for (i = paramsBegin; i != paramsEnd && * i->c_str() != '>'; ) {
      // set [parambegin,i) range 
      list<string>::iterator paramBegin = i;
      char c;
      while (i != paramsEnd, c = * i->c_str(), c != ',' && c != '>') {
	if (c == '(')
	  ForwardSkipBrackets(i, '(', ')', paramsEnd);
	else if (c == '<')
	  ForwardSkipBrackets(i, '<', '>', paramsEnd);
	else
	  ++i;
      }
      _templateParameters.push_back(MParameter(paramBegin, i, this));
      // skip ','
      //if (i != paramsEnd)
      if (* i->c_str() != '>')
	++i;
    }
    // End common part 
  }
  
  // now that the header is processed, load the directives
  LoadComments(cclass);

  // parse attributes and operations
  xmlNodePtr cchild;
  for (cchild = cclass->children; cchild != NULL; cchild = cchild->next) {
    if (cchild->type == XML_ELEMENT_NODE) {
      if (!xmlStrcmp(cchild->name, IS_UTF8 "var")) {
	MAttribute *mchild = new MAttribute(this);
	mchild->LoadParseTree(cchild);
	_children.insert(ChildrenMap::value_type(mchild->GetName(), mchild));
      }
      else if (!xmlStrcmp(cchild->name, IS_UTF8 "fct") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "clc") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "cld")) {
	MOperation *mchild = new MOperation(this);
	mchild->LoadParseTree(cchild);
	_children.insert(ChildrenMap::value_type(mchild->GetName(), mchild));
      }
      // nested classes 
      else if (!xmlStrcmp(cchild->name, IS_UTF8 "cls") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "str") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "tmc")) {
	/* Classes are allowed to be declared with qualified names but
	   we do not support it here. But neither does GCC so it doesn't
	   hurt much (yet) */
	string name = MClass::LoadName(cchild);
	ChildrenMap::iterator i;
	i = _children.find(name);
	if (i == _children.end()) {
	  MClass *mchild = new MClass(this);
	  _children.insert(ChildrenMap::value_type(name, mchild));
	  mchild->LoadParseTree(cchild, fileName);
	}
	else {
	  // TODO: Warning
	  cerr << "Duplicate class " << name << " ignored.\n";
	}
      }
      else if (!xmlStrcmp(cchild->name, IS_UTF8 "tokens") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "cmb") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "cmf") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "cmd")) {
	// skip, they are handled elsewhere
      }
      else if (
	       !xmlStrcmp(cchild->name, IS_UTF8 "enm") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "frc") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "frf") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "mac") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "mai") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "pri") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "pro") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "pub") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "typ") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "uni") ||
	       0) {
	// skip, we are not interested
      }
      else {
	// TODO Warning
	cerr << "Unknown element '" << (char*)cchild->name << "'.\n";
      }
    }
  }

}

void MClass::LoadUxf(xmlNodePtr uclass) {
  xmlNodePtr uchild;
  char *s;
  for (uchild = uclass->children; uchild != NULL; uchild = uchild->next) {
    if (uchild->type == XML_ELEMENT_NODE) {
      if (!xmlStrcmp(uchild->name, IS_UTF8 "Name")) {
	s = CAST_BAD xmlNodeGetContent(uchild);
	_name = s;
	free(s);
	// discard namespace prefixes
	string ns;
	while (StringPop(_name, "::", ns, _name))
	  ;
      }
      else if (!xmlStrcmp(uchild->name, IS_UTF8 "TaggedValue")) {
	string tag, value;
	GetTaggedValue(uchild, &tag, &value);
	if (tag == "DiagramPosition")
	  _diagramPosition.LoadString(value.c_str());
	else if (tag == "SourceFile")
	  _containingFiles.insert(value);
	else if (tag == "SourceLine")
	  _sourceInfo._sourceLine = atoi(value.c_str());
      }
    }
  }

  // Generalizations
  xmlXPathContextPtr ctx;
  xmlXPathObjectPtr obj;
  xmlNodeSetPtr nodes;
  string query;

  ctx = xmlXPathNewContext(uclass->doc);
  myXPathRegisterDocNs(ctx);
  ctx->node = uclass;

  query = "../uxf:Generalization[child::uxf:Child='" + GetQualifiedName() + "']";
  obj = xmlXPathEval(BAD_CAST query.c_str(), ctx);
  nodes = obj->nodesetval;
  for (int i = 0; i < nodes->nodeNr; ++i) {
    MGeneralization gen(this);
    gen.LoadUxf(nodes->nodeTab[i]);
    _superClasses.push_back(gen);
  }
  xmlXPathFreeObject(obj);

  //attributes, associations: in MModel, after all classes have been loaded

  xmlXPathFreeContext(ctx);
}

void MClass::LoadAssocAttrUxf(xmlNodePtr uDiamondAssocEnd) {
  MAttribute *mchild = new MAttribute(this);
  mchild->LoadUxf(uDiamondAssocEnd);
  string name = mchild->GetName();
  if (name.empty()) // this is the dummy AA of a one-way association
    delete mchild;
  else
    _children.insert(ChildrenMap::value_type(name, mchild));
}

void MClass::SaveUxf(xmlNodePtr /*ownerUxf*/, xmlNodePtr modelUxf) const {
  xmlNodePtr uclass;
  xmlNodePtr utparams;

  // oh, ugly...
  static int count;
  const_cast<MClass*>(this)->EnsureDiagramPosition(count);

  uclass = xmlNewNode(NULL, IS_UTF8 "Class");

  // (UXF: tagged values must come first)

  // insert a dummy to test the xslt code that takes it as input.
  // OK, it works.
  // AddTaggedValue(uclass, IS_UTF8 "Dummy", IS_UTF8 "Foo");

  // (this is a UXF extension)
  xmlAddChild(uclass, NewTaggedValue(IS_UTF8 "SourceFile", 
				     BAD_CAST SourceFile().c_str()));
  _sourceInfo.SaveUxfTo(uclass);

  // (this is a UXF extension)
  //if (!_diagramPosition.empty())
  xmlAddChild(uclass, NewTaggedValue(IS_UTF8 "DiagramPosition", 
				     BAD_CAST _diagramPosition.SaveString().c_str()));

  xmlNewChild(uclass, NULL, IS_UTF8 "Name", 
	      BAD_CAST GetQualifiedName().c_str());

  if (_isAbstract)
    xmlNewChild(uclass, NULL, IS_UTF8 "IsAbstract", IS_UTF8 "true");

  if (!_templateParameters.empty()) {
    utparams = xmlNewChild(uclass, NULL, IS_UTF8 "TemplateParameters", NULL);

    list<MParameter>::const_iterator i;
    for (i = _templateParameters.begin(); i != _templateParameters.end(); ++i) 
      xmlAddChild(utparams, i->SaveUxf()); 
  }

  // this class must be saved before its nested classes
  // so that it will exist when the nested class wants to be loaded into it
  xmlAddChild(modelUxf, uclass);

  ChildrenMap::const_iterator i;
  for (i = _children.begin(); i != _children.end(); ++i) 
    i->second->SaveUxf(uclass, modelUxf);

  {
    list<MGeneralization>::const_iterator i;
    for (i= _superClasses.begin(); i != _superClasses.end(); ++i)
      xmlAddChild(modelUxf, i->SaveUxf());
  }
}

void MClass::ParseDeclaration(list<string>::iterator b, list<string>::iterator e, 
			      list<string> &name, list<string> &type, list<string> &value, list<string> &params) {
  assert(b != e);

  typedef list<string>::iterator sli;
  // copy the tokens 
  sli i;
  for (i = b; i != e; ++i) {
    type.push_back(*i);
  }
  if (mDebug.IsEnabled('t')) {
    fprintf(stderr, " Input: "); 
    psl(type); 
    fprintf(stderr, "\n");
  }

  // find the name  
  // To appreciate what we need to deal with, first try understanding this:
  // void (*signal(int signum, void (*handler)(int)))(int)

  // positions, in order:
  sli beg = type.begin();    // beg=typeBegin        
  //  sli &typeBegin = beg;	     
  sli nameBegin;	     
  sli nameEnd;		     // nameEnd=paramBegin   
  sli &paramBegin = nameEnd; 
  // '(' ... ')'          
  sli paramEnd;
  // (const)
  // ( [] )               
  sli typeEnd;		     
  // (__attribute__)      
  // (=)                  
  sli valueBegin;
  // (0) 
  sli valueEnd;  
  sli end = type.end();	     

  // examples:
  //  int i;
  //  int x __attribute__ ((aligned (16))) = 0;
  //  int fun();
  //  int (fun)(); // same as above
  //  int (* iptr);

  valueEnd = end;
  --valueEnd; // points at the ";"
  if (* valueEnd != ";")
    ++valueEnd;

  sli lastParen = end;
  do {
    --lastParen;
  } while (lastParen != beg && *lastParen != ")");
  valueBegin = end;
  // this approach fails on "wxPrintData* data = (wxPrintData*) NULL"
  do {
    --valueBegin;
  } while (valueBegin != lastParen && *valueBegin != "=");
  bool isValueSpecified = (valueBegin != lastParen);
  if (!isValueSpecified)
    valueBegin = valueEnd;

  typeEnd = find(beg, end, "__attribute__");
  if (typeEnd == end) 
    typeEnd = valueBegin;

  if (isValueSpecified)
    ++valueBegin;
  // Now typeEnd, valueBegin and valueEnd are set.

  i = typeEnd;

  // skip "const" qualification of methods
  --i;
  if (*i != "const")
    ++i;

  if (PrevChar(i, beg)==')') {
    // Function parameters.
    paramEnd = i;
    BackSkipBrackets(i, '(', ')', beg);
    if (PrevChar(i, beg)==')') {
      --i; // Get inside the front pair of brackets
      if (PrevChar(i, beg)=='(') {
	// operator ()
	// pretend the parens are not there sothat they are a part of the name
	++i;
      }
      else {
	// They were parameters of a function pointed to.
	paramEnd = i;
	// Possibly the function pointer is returned by a function,
	// skip its parameters
	BackSkipBrackets(i, '(', ')', beg);
      }
    }
  }
  else {
    if (PrevChar(i, beg)==']') {
      // array
      do {
	BackSkipBrackets(i, '[', ']', beg);
      } while (PrevChar(i, beg)==']');
    }
    paramEnd = i;
  }
  nameEnd = i;
  // now done with arrays and function pointers  

  // mess with operators: operator ++, operator foo<bar>, operator char *
  // also destructors: ~ is separate
  assert (i != beg );
  --i;
  nameBegin = i;
  for (;;) {
    if (*i->c_str() == '~' || *i == "operator") 
      nameBegin = i;
    if (i == beg)    
      break;    
    --i;
  } 
  // the parameter name can be omitted. check if what we think is a name
  // really isn't a type, or a part of it.
  const char *s = nameBegin->c_str();
  if (*s == '*' || *s == '&' || *s == '.' || /* '.' '.' '.' */
      IsSimpleTypeSpecifier(*nameBegin) //||
      /* if the [parameter]name can be omitted */
      /*_scope->Has(*nameBegin)*/)
    ++nameBegin;


  // Now all is set, chop the results.

  //  name.splice(name.end(), type, nameBegin, nameEnd);
  name = list<string>(nameBegin, nameEnd);
  value = list<string>(valueBegin, valueEnd);
  params = list<string>(paramBegin, paramEnd);

  //last!
  type.erase(nameBegin, paramEnd);
  //  type = list<string>(typeBegin, typeEnd);
  type.erase(typeEnd, end);

  if (mDebug.IsEnabled('t')) {
    fprintf(stderr, " Name: %s\n", Detokenize(name).c_str());
    fprintf(stderr, " Type: %s\n", Detokenize(type).c_str());
    fprintf(stderr, " Value: %s\n", Detokenize(value).c_str());
    fprintf(stderr, " Params: %s\n", Detokenize(params).c_str());
    fprintf(stderr, "\n");
  }
}

void MClass::EnsureDiagramPosition(int &count) {
  if (_diagramPosition.IsUnspecified()) {
    int x = count % 10;
    int y = count;
    char buf[30];
    snprintf(buf, 29, "%d,%d", x, y);
    buf[29] = '\0';
    _diagramPosition.LoadString(buf);

    ++count;
  }
}

void MClass::Diff(const MModelElement &old, xmlNodePtr patchRoot) const {
  // the old element has to be of the same type
  const MClass & oldClass = dynamic_cast<const MClass &>(old);

  //xmlNodePtr patchNode;

  if (_diagramPosition != oldClass._diagramPosition) {
    string pos = _diagramPosition.SaveString();
    oldClass.DiffRequestChange(patchRoot, "@dia:pos", pos);
  }
  
  // diff superclasses
  list<MGeneralization>::const_iterator i;
  for (i = _superClasses.begin(); i != _superClasses.end(); ++i) {
    list<MGeneralization>::const_iterator j;
    // quadratic time, but most classes hava a single superclass anyway
    for (j = oldClass._superClasses.begin(); j != oldClass._superClasses.end(); ++j) {
      // coming from Uxf, their names are fully qualified
      if (i->_parent == j->_parent) {
	if (i->_route != j->_route) {
	  oldClass.DiffRequestChange(patchRoot, "@dia:route", 
				     i->_parent +";"+ i->_route.SaveString());
	}
      }
    }
  }
  
  // attributes, nested classes
  MNamespace::Diff(old, patchRoot);
}

void MClass::ProcessDirectives() {
  Directives::iterator i;
  for (i = _directives.begin(); i != _directives.end(); ++i) {
    string dname = i->first;
    string dvalue = i->second;

    if (dname == "@dia:pos") {
      _diagramPosition.LoadString(dvalue.c_str());
      if (mDebug.IsEnabled('c')) {
	fprintf(stderr," Diagram position:'%s'\n", 
		_diagramPosition.SaveString().c_str());
      }
    }
    else if (dname == "@dia:route") {
      string scName, route;
      string::size_type semipos = dvalue.find(';');
      if (semipos == string::npos) {
	route = dvalue;
      }
      else {
	scName = dvalue.substr(0, semipos);
	if (semipos < dvalue.size())
	  route = dvalue.substr(semipos + 1);
      }
      
      list<MGeneralization>::iterator i,j,e;
      i = _superClasses.begin();
      e = _superClasses.end();
      if (scName.empty()) {
	if (i == e) {
	  Warning("superfluous @dia:route, discarding");
	}
	else {
	  j = i;
	  ++i;
	  if (i == e) {
	    // only one superclass, route unnamed => ok
	    j->_route.LoadString(route.c_str());
	  }
	  else {
	    Warning("unassigned @dia:route, discarding");
	  }
	}
	  }
      else {
	while (i != e) {
	  MModelElement *parent = GetElement(i->_parent);
	  if (parent->GetQualifiedName() == scName) {
	    i->_route.LoadString(route.c_str());
	    break;
	  }
	  ++i;
	}
	if (i == e) {
	  Warning("misnamed @dia:route, discarding");
	}
      }
    } // if @dia:route
  } // for all directives

  // the same for the contained elements
  MNamespace::ProcessDirectives();
}

string MClass::SourceFile() const {
  return *_containingFiles.begin(); 
}
