/* $Id: diagram.cc,v 1.9 2002/03/17 19:09:36 martin Exp $ */
// Medoosa: Support for Documentation of C++ Programs
// Master's Thesis, Faculty of Mathematics and Physics, Charles University
// Martin Vidner, 2001
// Released under the terms of the GNU General Public License.
#include "diagram.h"
#include "common.h"
#include <math.h>

xmlNsPtr diaNS = NULL;

MDiagramPoint::MDiagramPoint() :
  _x(-1.0),
  _y(-1.0),
  _isUnspecified(true)
{
}

MDiagramPoint::MDiagramPoint(xmlNodePtr dnode) {
  LoadDia(dnode);
}

void MDiagramPoint::LoadDia(xmlNodePtr dnode) {
  xmlChar *val;
  val = xmlGetProp(dnode, IS_UTF8 "val");
  LoadString(CAST_BAD val);
  free(val);
}

void MDiagramPoint::LoadString(const char *s) {
  char *p;
  _x = strtod(s, &p);
  while (*p && *p != ',')
    ++p;
  if (*p == '\0') {
    _y = -1.0;
    _isUnspecified = true;
  }
  else {
    _y = strtod(p+1, NULL);
    _isUnspecified = false;
  }
}

xmlNodePtr MDiagramPoint::SaveDia() const {
  xmlNodePtr p;

  p = xmlNewNode(diaNS, IS_UTF8 "point");
  xmlSetProp(p, IS_UTF8 "val", BAD_CAST SaveString().c_str());
  
  return p;
}

string MDiagramPoint::SaveString() const {
  assert(!_isUnspecified);
  return FToS(_x) +","+ FToS(_y);
}

bool MDiagramPoint::operator == (const MDiagramPoint &p) const {
  const double eps = 1e-10;
  return !_isUnspecified && !p._isUnspecified &&
    fabs(_x - p._x) < eps && fabs(_y - p._y) < eps;
}

// ---------------------------------------------------------------

void MDiagramRoute::LoadDia(xmlNodePtr dnode) {
  xmlNodePtr node;
  for (node = dnode->children; node != NULL; node = node->next) {
    if (node->type == XML_ELEMENT_NODE) {
      _points.push_back(MDiagramPoint(node));
    }
  }
}

#define NEXT \
  do { ++s; } while (*s && *s != ','); \
  if (! *s++) return; \
  if (! *s) return
void MDiagramRoute::LoadString(const char *s) {
  bool vert; // is the current segment vertical?
  MDiagramPoint p;
  p._isUnspecified = false;

  if (! *s)
    return;

  vert = (*s == 'v');
  NEXT;

  p._x = atof(s);
  NEXT;
  p._y = atof(s);
  _points.push_back(p);
  NEXT;

  for (;;) {
    if (vert)
      p._y = atof(s);
    else
      p._x = atof(s);
    
    _points.push_back(p);
    vert = !vert;

    NEXT;
  }
}

xmlNodePtr MDiagramRoute::SaveDia() const {
  xmlNodePtr tv;
  xmlNodePtr val;
  xmlNodePtr points;
  xmlNodePtr orient;
  xmlNodePtr o;

  if (_points.empty()) {
    tv = xmlNewText(IS_UTF8 "");
  }
  else {
    tv = xmlNewNode(NULL, IS_UTF8 "TaggedValue");
    xmlNewChild(tv, NULL, IS_UTF8 "Tag", IS_UTF8 "DiagramRoute");
    val = xmlNewChild(tv, NULL, IS_UTF8 "Value", NULL);

    points = xmlNewChild(val, diaNS, IS_UTF8 "attribute", NULL);
    xmlSetProp(points, IS_UTF8 "name", IS_UTF8 "orth_points");
    orient = xmlNewChild(val, diaNS, IS_UTF8 "attribute", NULL);
    xmlSetProp(orient, IS_UTF8 "name", IS_UTF8 "orth_orient");

    list<MDiagramPoint>::const_iterator i, j, e;
    i = _points.begin();
    e = _points.end();

    xmlAddChild(points, i->SaveDia());
    j = i;
    ++i;
    assert(i != e);
    do {
      xmlAddChild(points, i->SaveDia());
      o = xmlNewChild(orient, diaNS, IS_UTF8 "enum", NULL);
      xmlSetProp(o, IS_UTF8 "val", IS_UTF8 (i->_x == j->_x? "1":"0"));
	
      j = i;
      ++i;
    } while (i != e);
  }
  return tv;
}

string MDiagramRoute::SaveString() const {
  list<MDiagramPoint>::const_iterator p, q, e;
  string out;
  e = _points.end();

  q = _points.begin();
  assert(q != e);

  p = q;
  ++q;
  assert(q != e);

  if (p->_y != q->_y) {
    // first segment is vertical
    out += "v";
    out += ",";
    out += FToS(p->_x);
    out += ",";
    out += FToS(p->_y);

    p = q;
    ++q;
    
    out += ",";
    out += FToS(p->_y);
  }
  else {
    // first segment is horizontal
    out += "h";
    out += ",";
    out += FToS(p->_x);
    out += ",";
    out += FToS(p->_y);
  }
  
  while (q != e) {
    // *p to *q is horizontal, p->_x has been saved
    assert(p->_y == q->_y);
    out += ",";
    out += FToS(q->_x);
    p = q;
    ++q;

    if (q != e) {
      // *p to *q is vertical, p->_y has been saved
      out += ",";
      out += FToS(q->_y);
      p = q;
      ++q;
    }
  }
  return out;
}

// ---------------------------------------------------------------

MAssociationRoute::MAssociationRoute() {
  _connectionPoints[0] = _connectionPoints[1] = -1;
}

void MAssociationRoute::LoadString(const string &s) {
  string connectionPoints, cp1, cp2, route;
  StringPop(s, ";", connectionPoints, route);
  if (connectionPoints.empty()) {
    // TODO
    cerr << "Input error, route conection points not specified\n";
  }
  else {
    StringPop(connectionPoints, ",", cp1, cp2);
    if (cp1.empty() || cp2.empty()) {
      // TODO
      cerr << "Input error, no comma to separate route connection points\n";
    }
    else {
      _connectionPoints[0] = atoi(cp1.c_str());
      _connectionPoints[1] = atoi(cp2.c_str());
    }
  }

  MDiagramRoute::LoadString(route.c_str());  
}

string MAssociationRoute::SaveString() const {
  assert(_connectionPoints[0] >= 0);
  return DToS(_connectionPoints[0]) + "," + DToS(_connectionPoints[1]) + ";" +
    MDiagramRoute::SaveString();
}

void MAssociationRoute::SaveDia(xmlNodePtr uassoc) const {
  if (_connectionPoints[0] >= 0) {
    xmlAddChild(uassoc,
		NewTaggedValue(IS_UTF8 "DiagramConnection1",
			       BAD_CAST DToS(_connectionPoints[0]).c_str()));
    xmlAddChild(uassoc,
		NewTaggedValue(IS_UTF8 "DiagramConnection2",
			       BAD_CAST DToS(_connectionPoints[1]).c_str()));
  }
  xmlAddChild(uassoc, MDiagramRoute::SaveDia());
}
