// $Id: namespace.cc,v 1.13 2002/03/17 19:09:36 martin Exp $
// Medoosa: Support for Documentation of C++ Programs
// Master's Thesis, Faculty of Mathematics and Physics, Charles University
// Martin Vidner, 2001
// Released under the terms of the GNU General Public License.
#include "namespace.h"
#include <libxml/xpath.h>
#include <iostream>
#include <typeinfo>
#include "class.h"
#include "debug.h"
#include "common.h"

void MNamespace::DeleteChildren() {
  ChildrenMap::iterator i;

  for (i = _children.begin(); i != _children.end(); ++i) 
    delete i->second;
  _children.clear();
}

void MNamespace::CopyChildren(const MNamespace & a) {
  MModelElement *e;
  ChildrenMap::const_iterator i;
  for (i = a._children.begin(); i != a._children.end(); ++i) {
    e = i->second->Clone();
    e->SetOwner(this);
    _children.insert(_children.end(), ChildrenMap::value_type(i->first, e));
  }
}

MNamespace::~MNamespace() {
  DeleteChildren();
}

MNamespace::MNamespace(const MNamespace &a) :
  MModelElement(a)
{
  CopyChildren(a);
}

  
MNamespace & MNamespace::operator = (const MNamespace &a) {
  if (&a != this) {
    MModelElement::operator = (a);

    DeleteChildren();
    CopyChildren(a);
  }
  return *this;
}

void MNamespace::SaveUxf(xmlNodePtr ownerUxf, xmlNodePtr modelUxf) const {
  // just pass it on
  ChildrenMap::const_iterator i;
  for (i = _children.begin(); i != _children.end(); ++i) 
    i->second->SaveUxf(ownerUxf, modelUxf);
}

void MNamespace::ProcessDirectives() {
  // just pass it on
  ChildrenMap::iterator i;
  for (i = _children.begin(); i != _children.end(); ++i) 
    i->second->ProcessDirectives();
}

string MCxxNamespace::LoadName(xmlNodePtr cnode) {
  list<string> toks = LoadTokens(FindFirstTokenChild(cnode));
  list<string>::iterator i = toks.begin();
  ++i;
  assert(i != toks.end());
  return *i;
}

void MNamespace::Diff(const MModelElement &old, xmlNodePtr patchRoot) const {
  // the old element has to be of the same type
  const MNamespace & oldNs = dynamic_cast<const MNamespace &>(old);
  ChildrenMap::const_iterator i;
  ChildrenMap::const_iterator oldi;
  for (i = _children.begin(); i != _children.end(); ++i) {
    const MModelElement *e = i->second;
    string name = e->GetName();
    if (mDebug.IsEnabled('d'))
      cerr << "New elem: " << e->GetQualifiedName() << endl;

    oldi = oldNs._children.find(name);
    if (oldi == oldNs._children.end()) {
      cerr << "Not found in original, discarded: " << name << ".\n";
      // TODO: detect also Is in original but not in new
    }
    else {
      const MModelElement *olde = oldi->second;
      if (typeid(*olde) == typeid(*e)) {
	e->Diff(*olde, patchRoot);
      }
      else {
	cerr << "Incompatible types: "
	     << "new '" << typeid(*e).name() << "', "
	     << "old '" << typeid(*olde).name() << "'.\n";
      }
    }
  }
}

MModelElement * MNamespace::GetElementElsewhere(const string &name) const {
  //  TODO: try base classes.
  // pass it up
  if (_owner != NULL) {
    return _owner->GetElement(name);
  }
  else { // I am the root
    // and cannot find it
    return NULL;
  }
}

MModelElement * MNamespace::GetElement(const string &qname) const {
  // Each recursive call either goes up one namespace or strips the
  // leading namespace off the qname, therefore the recursion is finite.

  //TODO: use "using"

  string nsName, tail;
  if (StringPop(qname, "::", nsName, tail)) { // qualified name

    if (nsName.empty()) { // root scope: ::foo..., pass it up
      if (_owner != NULL) {
	return _owner->GetElement(qname);
      }
      else { // I am the root
	// this recursion sucks
	return this->GetElement(tail);
      }
    }
    else { // use a namespace inside
      ChildrenMap::const_iterator nsi = _children.find(nsName);
      if (nsi != _children.end()) {  // there's something of that name
	MModelElement * nse = nsi->second;
	MNamespace * ns = dynamic_cast<MNamespace *>(nse);
	if (ns != NULL) { // and it is a namespace indeed
	  return ns->GetElement(tail);
	}
	else { // it is not a namespace, the program is ill-formed
	  return NULL;
	}
      }
      else { // not found here
	return GetElementElsewhere(qname);
      }
    }
  }
  else { // unqualified name
    ChildrenMap::const_iterator ei = _children.find(tail);
    if (ei != _children.end()) { // glory, we found it!
      // TODO: luckily we don't do overload resolution
      return ei->second; 
    }
    else { // not found here
      return GetElementElsewhere(tail);
    }
  }
  assert(0); //not reached
  return NULL;
}

MClass* MNamespace::InsertQualifiedClass(const string& qname) {
  MClass *result;

  string nsName, tail;
  if (StringPop(qname, "::", nsName, tail)) { // qualified name
    if (nsName.empty()) { // root scope: ::foo..., pass it up
      //^ this should never happen, but it does not hurt 
      // to have it in sync with GetElement 
      if (_owner != NULL) {
	result = _owner->InsertQualifiedClass(qname);
      }
      else { // I am the root
	// this recursion sucks
	result = this->InsertQualifiedClass(tail);
      }
    }
    else { // use/create a namespace inside
      ChildrenMap::const_iterator nsi = _children.find(nsName);
      MNamespace *ns;
      if (nsi != _children.end()) {  // there's something of that name
	MModelElement * nse = nsi->second;
	ns = dynamic_cast<MNamespace *>(nse);
	if (ns == NULL) { // it is not a namespace, the model is ill-formed
	  cerr << "Not a class or namespace: "<< nsName 
	       << ", cannot create " << tail << ".\n";
	  return NULL;
	}
      }
      else { // not found, need to create it
	ns = new MCxxNamespace(nsName, this);
	_children.insert(ChildrenMap::value_type(nsName, ns));
      }
      result = ns->InsertQualifiedClass(tail);
    }
  }
  else { // unqualified name
    result = new MClass(this);
    _children.insert(ChildrenMap::value_type(tail, result));
  }
  return result;
}

// ---------------------------------------------------------------

MCxxNamespace::~MCxxNamespace() {
}

void MCxxNamespace::LoadParseTree(xmlNodePtr cnode, const string &fileName) {
  MNamespace * mchild;
  xmlNodePtr cchild;

  _containingFiles.insert(fileName);

  if (!IsRoot())
    _name = LoadName(cnode);

  for (cchild = cnode->children; cchild != NULL; cchild = cchild->next) {
    if (cchild->type == XML_ELEMENT_NODE) {
      string name;

      if (!xmlStrcmp(cchild->name, IS_UTF8 "cls") ||
	  !xmlStrcmp(cchild->name, IS_UTF8 "str") ||
	  !xmlStrcmp(cchild->name, IS_UTF8 "tmc")) {
	/* Classes are allowed to be declared with qualified names but
	   we do not support it here. But neither does GCC so it doesn't
	   hurt much (yet) */
	// extend it so that it returns a qname
	name = MClass::LoadName(cchild); 
	ChildrenMap::iterator i;
	// if using InsertQualifiedClass later,
	// use GetElement here. check that it works.
	i = _children.find(name); 
	if (i == _children.end()) {
	  // InsertQualifiedClass is a special case, might work or not.
	  // mchild = InsertQualifiedClass(name);
	  mchild = new MClass(this);
	  _children.insert(ChildrenMap::value_type(name, mchild));
	  mchild->LoadParseTree(cchild, fileName);
	}
	else {
	  cerr << "Duplicate class " << name << " ignored.\n";
	}
      }
      else if (!xmlStrcmp(cchild->name, IS_UTF8 "nms")) {
	name = MCxxNamespace::LoadName(cchild);

	ChildrenMap::iterator i;
	i = _children.find(name);
	if (i == _children.end()) {
	  // first occurrence of this namespace
	  mchild = new MCxxNamespace(name, this);
	  _children.insert(ChildrenMap::value_type(name, mchild));
	}
	else {
	  // adding to a previously defined namespace
	  mchild = dynamic_cast<MNamespace*>(i->second);
	}
	mchild->LoadParseTree(cchild, fileName);
	

	// ... solve unnamed root
      }
      else if (!xmlStrcmp(cchild->name, IS_UTF8 "tokens") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "cmb") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "cmf") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "cmd")) {
	// skip, they are handled elsewhere
      }
      else if (
	       !xmlStrcmp(cchild->name, IS_UTF8 "cfw") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "enm") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "fct") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "inc") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "mac") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "mai") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "scf") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "stf") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "typ") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "uni") ||
	       !xmlStrcmp(cchild->name, IS_UTF8 "var") ||
	       0) {
	// skip, we are not interested
      }
      else {
	cerr << "Unknown element '" << (char*)cchild->name << "'.\n";
      }
    }
  }
}

void MCxxNamespace::LoadUxf(xmlNodePtr unode) {
  xmlXPathContextPtr ctx;
  xmlXPathObjectPtr obj;
  xmlNodeSetPtr nodes;
  xmlNodePtr uclass;
  int i;
  MClass *cl;

  ctx = xmlXPathNewContext(unode->doc);
  myXPathRegisterDocNs(ctx);
  ctx->node = unode;
  obj = xmlXPathEval(IS_UTF8 "uxf:Class", ctx);
  nodes = obj->nodesetval;

  for (i = 0; i < nodes->nodeNr; ++i) {
    uclass = nodes->nodeTab[i];

    xmlXPathContextPtr classCtx;
    xmlXPathObjectPtr nameObj;
    string name;

    classCtx = xmlXPathNewContext(uclass->doc);
    myXPathRegisterDocNs(classCtx);

    classCtx->node = uclass;
    nameObj = xmlXPathEval(IS_UTF8 "string(uxf:Name)", classCtx);
    name = CAST_BAD nameObj->stringval;
    if (mDebug.IsEnabled('d'))
      cerr << name << endl;

    cl = InsertQualifiedClass(name);
    cl->LoadUxf(uclass);

    xmlXPathFreeObject(nameObj);
    xmlXPathFreeContext(classCtx);
  }

  xmlXPathFreeObject(obj);
  xmlXPathFreeContext(ctx);
}
