/* $Id: patch-ccdoc.cc,v 1.11 2001/04/23 16:10:30 martin Exp $ */
// Medoosa: Support for Documentation of C++ Programs
// Master's Thesis, Faculty of Mathematics and Physics, Charles University
// Martin Vidner, 2001
// Released under the terms of the GNU General Public License.
#include <stdlib.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#define IS_UTF8 (xmlChar *)
#define CAST_BAD (char *)
#include <string>
#include <list>
#include <vector>
#include <map>
#include <fstream>
#include <regex.h>
#include "debug.h"

/** For an element (eg. class, attribute) maps directive names to
 values.

 If the value is an empty string, the directive will bw
 deleted.  So, valueless directives, such as @uml:aggreg, have to
 specify a space " " as the value. */
typedef map<string, string> MElementDirectiveMap;

/** Describes patches to an  element.
 *
 * Before the directives are patched, a check is made that the patch
 * is not misplaced: shortName is searched for in the line and if not
 * found, a warning is issued using the fullName. */
struct MElementPatches {
  string _fullName;
  string _shortName;
  MElementDirectiveMap _directives;
};

/** for a file, maps line numbers to elements */
typedef map<int, MElementPatches> MFileDirectiveMap;

/** for a project (diagram?, package?), maps file names to files */
typedef map<string, MFileDirectiveMap> MProjectDirectiveMap;

MProjectDirectiveMap projectMap;

regex_t rx_directive;
 /** @sample:a-directive 1,1 */
char *pat_directive = "^[ \t]*/\\*\\*[ \t]*(@[A-Za-z:_-]+).*\\*/[ \t]*\n$";

void LoadProjectMap(MProjectDirectiveMap &pm, xmlDocPtr patchDoc) {
  xmlNodePtr root = xmlDocGetRootElement(patchDoc);
  xmlNodePtr request;
  xmlNodePtr child;

  for (request = root->children; request != NULL; request = request->next) {
    if (request->type == XML_ELEMENT_NODE) {
      string file, fullName, shortName, directive, value;
      int line;


      for (child = request->children; child != NULL; child = child->next) {
	if (child->type == XML_ELEMENT_NODE) {
	  char *s;
	  s = CAST_BAD xmlNodeGetContent(child);
	
	  if (!xmlStrcmp(child->name, IS_UTF8 "file"))
	    file = s;
	  else if (!xmlStrcmp(child->name, IS_UTF8 "line"))
	    line = atoi(s);
	  else if (!xmlStrcmp(child->name, IS_UTF8 "full"))
	    fullName = s;
	  else if (!xmlStrcmp(child->name, IS_UTF8 "short"))
	    shortName = s;
	  else if (!xmlStrcmp(child->name, IS_UTF8 "directive"))
	    directive = s;
	  else if (!xmlStrcmp(child->name, IS_UTF8 "value"))
	    value = s;
	  free(s);
	}
      }
      
      MElementPatches & ep = pm[file][line];
      if (!ep._fullName.empty() && ep._fullName != fullName) {
	cerr << "Warning, element mismatch for " << file << ':' << line
	     << ", " << fullName << " != " << ep._fullName << endl;
      }
      if (!ep._shortName.empty() && ep._shortName != shortName) {
	cerr << "Warning, element mismatch for " << file << ':' << line
	     << ", " << shortName << " != " << ep._shortName << endl;
      }
      ep._fullName = fullName;
      ep._shortName = shortName;
      ep._directives[directive] = value;
    }
  }
}

void DoDirectives(MElementPatches &ep,  vector<string> &lines, int line) {
  vector<string>::iterator element_li(lines.begin() + line - 1);

  string element_s = *element_li;
  // check that it is the correct line
  string::size_type pos = element_s.find(ep._shortName);
  if (pos == string::npos) {
    cerr << "Warning, " << ep._shortName 
	 << " (" << ep._fullName <<") not found at line " << line << endl;
  }

  MElementDirectiveMap &em = ep._directives;
  string indent(element_s, 0, element_li->find_first_not_of(" \t"));

    // DUH, match[0] is the whole matched string, I want match[1] !
  regmatch_t match[2];
  vector<string>::reverse_iterator li(element_li);
  while (li != lines.rend() &&
	 regexec(&rx_directive, li->c_str(), 2, match, 0) != REG_NOMATCH) {
    string directive(*li, match[1].rm_so, match[1].rm_eo - match[1].rm_so);
    if (mDebug.IsEnabled('p')) {
      cerr << "Matched " 
	 << match[0].rm_so << ','
	 << match[0].rm_eo << ' '
	 << match[1].rm_so << ','
	 << match[1].rm_eo << ' '
	 << '\'' << directive << "'\n";
    }

    // is there a request to patch this directive?
    MElementDirectiveMap::iterator ei = em.find(directive);
    if (ei != em.end()) {
      string value = ei->second;
      if (value.empty()) {
	// delete the directive, that is, empty the string but don't
	// erase the line from the vector
	//	li->clear();
	*li = "";
      }
      else {
	// the directive in the file has to be patched
	*li = indent + "/** " + directive + " " + value + " */\n";
      }
      // done with this one
      em.erase(ei);
    }

    ++li;
  }
   
  // now put in the remaining patch requests for the directives that
  // weren't in the file;
  string rest;
  MElementDirectiveMap::iterator ei;
  for (ei = em.begin(); ei != em.end(); ++ei) {
    string value = ei->second;
    if (!value.empty())
      rest += indent + "/** " + ei->first + " " + value + " */\n";
  }
  *element_li = rest + *element_li;  
}

void DoLines(MFileDirectiveMap &fm, vector<string> &lines) {
  MFileDirectiveMap::iterator i;
  for (i = fm.begin(); i != fm.end(); ++i) {
    int line = i->first;
    MElementPatches &ep = i->second;
    DoDirectives(ep, lines, line);
  }
}

void DoFiles(MProjectDirectiveMap &pm) {
  MProjectDirectiveMap::iterator i;
  for (i = pm.begin(); i != pm.end(); ++i) {
    string file = i->first;
    MFileDirectiveMap & fm = i->second;

    cerr << "Patching " << file << ".\n";

    ifstream is(file.c_str());
    string buf;
    // the lines include the newline
    vector<string> lines;
    //int i = 0;
    for (;;) {
      //++i;
      getline(is, buf);

      if (!is && buf.empty())
	break;

      if (mDebug.IsEnabled('P'))
	cerr << '"' << buf << '"' << endl;
      buf += '\n';
      lines.push_back(buf);
    }

    DoLines(fm, lines);

    // make a backup
    string bfile = file + '~';
    if (rename(file.c_str(), bfile.c_str()) != 0) {
      perror("Error backing up");
    }

    // output
    ofstream os(file.c_str());    
    vector<string>::iterator si;
    for (si = lines.begin(); si != lines.end(); ++si) {
      os << *si;
    }  
  }
}

int main(int argc, char **argv) {
  xmlDocPtr patchDoc;
  int result;

  if (getenv("MEDOOSA_DEBUG_ALL") != NULL)
    mDebug.EnableAll();
  else
    mDebug.Enable(getenv("MEDOOSA_DEBUG"));

  if (argc != 1) {
    fprintf(stderr, "usage: %s\n", argv[0]);
    exit(1);
  }
     
  result = regcomp(&rx_directive, pat_directive, REG_EXTENDED);
  if (result) {
    char buf[100];
    regerror(result, &rx_directive, buf, 100);
    fprintf(stderr, "regcomp error: %s\n", buf);
    exit(1);
  }

  LIBXML_TEST_VERSION;
  patchDoc = xmlParseFile("-");
  if (patchDoc == NULL)
    exit(1);

  LoadProjectMap(projectMap, patchDoc);
  DoFiles(projectMap);  

  xmlFreeDoc(patchDoc);

  regfree(&rx_directive);
  return 0;
}
