// -*- C++ -*-
/* $Id: type.cc,v 1.11 2001/04/23 16:10:30 martin Exp $ */
// Medoosa: Support for Documentation of C++ Programs
// Master's Thesis, Faculty of Mathematics and Physics, Charles University
// Martin Vidner, 2001
// Released under the terms of the GNU General Public License.

#include <typeinfo>
#include "type.h"
#include "debug.h"
#include "class.h"

bool IsSimpleTypeSpecifier(const string &s) {
  return s == "char" || s == "wchar_t" || s == "bool" || 
    s == "short" || s == "int" || s == "long" || 
    s == "signed" || s == "unsigned" || 
    s == "float" || s == "double" || s == "void";
}


MType::~MType() {
  if (_parsed != NULL)
    delete _parsed;
}

MType::operator const list<string> () const {
  return _tokens;
}

xmlNodePtr MType::SaveUxf(const xmlChar *tagname) const {
  xmlNodePtr utype;
  utype = xmlNewNode(NULL, tagname);

  /*  
  list<string>::iterator i;
  for (i = _tokens.begin(); i != _tokens.end(); ++i) {
    xmlNodeAddContent(utype, BAD_CAST i->c_str());
    xmlNodeAddContent(utype, IS_UTF8 " ");
  }
  */
  xmlNodeAddContent(utype, BAD_CAST Detokenize(_tokens).c_str());
  return utype;
}

MTypeAnalysis MType::Analyze() {
  if (_parsed == NULL)
    ParseType();

  MTypeAnalysis a;
  if (_parsed == NULL)
    cerr << "Empty type encountered.\n";
  else
    _parsed->Analyze(a, _ns);
  return a;
}

void MType::ParseType() {
  //assert all classes loaded

  list<string>::iterator i = _tokens.begin();
  _parsed = ParseTypeRecursive(i);
}

MTypeComponent* MType::ParseTypeRecursive(list<string>::iterator &cursor) {
  MTypeComponent *result = NULL;

  while (cursor != _tokens.end()) {
    if (*cursor == "&") {
      result = new MTCReference(result);
    }
    else if (*cursor == "*") {
      assert(result != NULL);
      MTCNameSpecifier *q = dynamic_cast<MTCNameSpecifier*>(result);
      if (q != NULL) {
	return new MTCMemberPointer(q);
      }
      else {
	result = new MTCPointer(result);
      }
    }
    else if (*cursor == "::") {
      if (result == NULL)
	result = new MTCNameSpecifier(NULL);
      else {
	MTCClass *c = dynamic_cast<MTCClass*>(result);
	assert(c != NULL);
	result = new MTCNameSpecifier(c);
      }
    }
    else if (*cursor == "<") {
      MTCClass *c = dynamic_cast<MTCClass*>(result);
      MTCTemplateClass *t = dynamic_cast<MTCTemplateClass*>(result);
      assert(c != NULL && t == NULL);
      t = new MTCTemplateClass(c);
      ++cursor;
      for (;;) {
	MTypeComponent *param = ParseTypeRecursive(cursor);
	t->AddParam(param);
	if (cursor ==_tokens.end() || *cursor == ">")
	  break;
	++cursor;
      }
      result = t;
    }
    else if (*cursor == ">" || *cursor == ",") {
      // template parameter
      return result;
    }
    else if (*cursor == "[") {
      list<string>::iterator bbegin, bend;
      ++cursor; 
      bbegin = cursor; 
      --cursor;

      ForwardSkipBrackets(cursor, '[', ']', _tokens.end());

      --cursor; 
      bend = cursor; 
      result = new MTCArray(result, Detokenize(list<string>(bbegin, bend)));
    }
    else if (*cursor == "]") {
      cerr << "Closing bracket cannot happen!\n";
    }
    else if (*cursor == "(") {
      ++cursor;
      if (*cursor == "*") {
	// function pointers are a "short circuit"
	cursor = _tokens.end();
	return new MTCFunctionPointer(result);
      }
      else {
	// otherwise, sudden death
	cerr << "Cannot handle a parenthesis\n";
	--cursor;
	ForwardSkipBrackets(cursor, '(', ')', _tokens.end());
	--cursor;
      }
    }
    else if (*cursor == ")") {
      cerr << "Closing parenthesis cannot happen!\n";
    }
    else if (*cursor == "const" || *cursor == "volatile") {
      // do nothing, Changeability not supported.
      // but if it ever is, remember to handle "const Foo" as "Foo const" 
    }
    else if (*cursor == "typename" || *cursor == "template" ||
	     *cursor == "class" || *cursor == "struct" || 
	     *cursor == "union" || *cursor == "enum" ) {
      // skip
    }
    else if (IsSimpleTypeSpecifier(*cursor)) {
      if (result == NULL) {
	result = new MTCFundamental(*cursor);
      }
      else {
	MTCFundamental *fresult = dynamic_cast<MTCFundamental*>(result);
	assert(fresult != NULL);
	fresult->Append(*cursor);
      }
    }
    else /*if (it is a class name)*/ {
      if (result == NULL) {
	result = new MTCClass(*cursor);
      }
      else {
	MTCNameSpecifier *q = dynamic_cast<MTCNameSpecifier*>(result);
	if (q != NULL) {
	  result = new MTCClass(*cursor, q);
	}
	else {
	  if (mDebug.IsEnabled('t'))
	    cerr << "Member pointer base begins: " << *cursor << endl;
	  
	  MTypeComponent *tc = ParseTypeRecursive(cursor);
	  MTCMemberPointer *mptr = dynamic_cast<MTCMemberPointer*>(tc);
	  if (mptr != NULL) {
	    mptr->SetTarget(result);
	    result = mptr;
	  }
	  else {
	    cerr << "It was not a member pointer.\n";
	    MTCClass *c = dynamic_cast<MTCClass*>(result);
	    cerr << "Discarding previous component ";
	    if (c != NULL)
	      cerr << c->GetName() << endl;
	    else
	      cerr << typeid(*tc).name() << endl;

	    delete result;
	    result = tc;
	  }
	}
      }      
    } //if

    if (cursor != _tokens.end())
      ++cursor;
  } // while
  return result;
}

// ---------------------------------------------------------------

void MTCFundamental::Analyze(MTypeAnalysis &, MNamespace *) {
  // do nothing, defaults are ok
}

// ---------------------------------------------------------------

MTCNameSpecifier::MTCNameSpecifier(MTCClass *aclass) : 
  _name(aclass->GetName()+"::") 
{ 
  delete aclass; 
}


// ---------------------------------------------------------------

void MTCClass::Analyze(MTypeAnalysis &result, MNamespace *ns) {
  MClass *cls = dynamic_cast<MClass*>(ns->GetElement(_name));
  // if we can connect to it
  // TODO use class stub
  if (cls != NULL) {
    // TODO use traits
    result._targetType = _name;
    result._aggregation = COMPOSITE;
    result._multiplicity.LoadString("1");
  }
}

MTCClass::MTCClass(const string &name, MTCNameSpecifier *q) { 
  _name = q->GetName() + name; 
  delete q; 
}

// ---------------------------------------------------------------

MTCTemplateClass::~MTCTemplateClass() {
  list<MTypeComponent *>::iterator i;
  for (i = _parameters.begin(); i != _parameters.end(); ++i)
    delete *i;
}

void MTCTemplateClass::Analyze(MTypeAnalysis &result, MNamespace *ns) {
  MClass *cls = dynamic_cast<MClass*>(ns->GetElement(_name));
  // if we can connect to it
  // TODO use class stub
  if (1 || cls != NULL) {
    // if (!cls->IsTemplate())...
    // TODO use traits
    // this is a heuristic, works with STL containers
    list<MTypeComponent *>::reverse_iterator tci = _parameters.rbegin();
    list<MTypeComponent *>::reverse_iterator re = _parameters.rend();
    if (tci == re) {
      // no template parameters? maybe defaults used, makes sense to use
      // our class nature
      MTCClass::Analyze(result, ns);
    }
    else {
      // the last parameter is the target
      (*tci)->Analyze(result, ns);

      if (!result._targetType.empty()) {
	MMultiplicity star;
	star.LoadString("*");
	result._multiplicity.Compose(star);
	result._aggregation = AGGREGATE;

	// check for qualifier, which is the penultimate
	++tci;
	if (tci != re) {
	  // DUH, need MTC::stringize. that's too much at this time,
	  // let's default on @uml:qualif
	}
      }
    }
  }
}

// ---------------------------------------------------------------

void MTCReference::Analyze(MTypeAnalysis &result, MNamespace *ns) {
  _target->Analyze(result, ns);
  // cancel COMPOSITE:
  result._aggregation = NONE;
}

// ---------------------------------------------------------------

void MTCPointer::Analyze(MTypeAnalysis &result, MNamespace *ns) {
  _target->Analyze(result, ns);
  // cancel COMPOSITE:
  result._aggregation = NONE;
  // clear multiplicity
  result._multiplicity.LoadString("");
}

// ---------------------------------------------------------------

void MTCMemberPointer::Analyze(MTypeAnalysis &, MNamespace *) {
  // do nothing, defaults are ok
}

// ---------------------------------------------------------------

MTCArray::MTCArray(MTypeComponent *target, const string& bound) :
  _target(target)
{
  const char *b = bound.c_str();
  char *e;
  strtol(b, &e, 10);
  if (*e == '\0') {
    // a number
    _multiplicity.LoadString("0.." + bound);
  }
  else {
    // a constant, MMultiplicity is too dumb
    _multiplicity.LoadString("*");
  }
}

void MTCArray::Analyze(MTypeAnalysis &result, MNamespace *ns) {
  _target->Analyze(result, ns);
  result._multiplicity.Compose(_multiplicity);
}

// ---------------------------------------------------------------

void MTCFunctionPointer::Analyze(MTypeAnalysis &, MNamespace *) {
  // do nothing, defaults are ok
}

