// -*- C++ -*-
/* $Id: type.h,v 1.18 2001/04/23 16:10:30 martin Exp $ */
// Medoosa: Support for Documentation of C++ Programs
// Master's Thesis, Faculty of Mathematics and Physics, Charles University
// Martin Vidner, 2001
// Released under the terms of the GNU General Public License.
#ifndef __M_TYPE_H
#define __M_TYPE_H

#include <list>
#include <string>
#include "common.h"
#include "umltypes.h"
#include "element.h"

bool IsSimpleTypeSpecifier(const string &s);


class MTypeAnalysis;
class MTypeComponent;

/** @dia:pos -162,35 */
class MType {
public:
  MType(MNamespace *ns) : _ns(ns), _parsed(NULL) {}
  ~MType();
  /** Construct. The input range includes a name that has the type.
      That name will be excluded from the type. */
  //  void Load(list<string> tokens, list<string>::iterator b, list<string>::iterator e);
  void Load(const list<string> &tokens) { _tokens = tokens; }
  /** converts */
  operator const list<string> () const;
  xmlNodePtr SaveUxf(const xmlChar *tagname) const;

  /** Describes itself */
  MTypeAnalysis Analyze();
  
private:
  /** @dia:route 12,13;h,12,26.7,5,2,-128.234,38.7 */
  MNamespace *_ns;
  list<string> _tokens;
  /** @dia:route 5,1;h,-162,44,-152,53,-144.056,55 */
  /** @uml:mul 1 */
  MTypeComponent *_parsed;

  /** Converts _tokens, a string list, to _parsed, a tree */
  void ParseType();
  MTypeComponent* ParseTypeRecursive(list<string>::iterator &cursor);

  /*TODO
  // prohibit
  MType(const MType&);
  MType& operator= (const MType&);
  */
};

// ---------------------------------------------------------------

// maybe differentiate InternalME (typedef, classstub: no
// load/saveuxf/diff) and ExternalME
// actually, IME is not a ME at all, lets have a NamedElement as the root?
/** @dia:pos -51,27 */
/** @dia:route MModelElement;v,-17.2072,17.8,20,-44.0552,27 */
class MTypedef : public MModelElement {
  MTypedef(MNamespace *ns) : MModelElement("", ns), _type(ns) {} //OK?
private:
  /** @dia:route 4,0;h,-128.234,35.7,-108,27,-51 */
  MType _type;
};

// ---------------------------------------------------------------

/*
Big picture: 

A C++ attribute (data member) has a type and we want to analyze the
type to determine whether the attribute should be represented as a UML
attribute or as an end of a UML association. We also want the details
such as multiplicity ((changeability)) and, in the latter case,
whether the association end is qualified, whether it is a composite or
an aggregate and what kind of ordering the links must satisfy.

Ccdoc supplies the type as a list of tokens so unless the analysis is
going to be a dirty hack, the syntax has to be properly parsed. The
result is a tree of type components (MTypeComponent) [Describe it. Examples?]

MTypeComponent's Analyze method recursively descends the tree to store
its findings in a MTypeAnalysis structure. [Describe the struct,
example] This structure and the user supplied @-directives for the
data member then determine the result.



TODO how MTC::Analyze uses *Traits

*/

/** Keeps information about a C++ type, specifically, which UML-like
 * constructs it models.  */
/** @dia:pos -120,48 */
class MTypeAnalysis {
public:
  MTypeAnalysis() : _aggregation(NONE), _ordering(UNORDERED) {}
  string _targetType; //TODO: (q)name
  /** @dia:route 8,2;h,-95,53.9,-100,48,-104.171 */
  MMultiplicity _multiplicity;
  // _changeability;

  // The following apply only to an association end, not an attribute:
  list<string> _qualifierTypes; //TODO: (q)name
  // duh, this is wrong, _obviously_ aggregation!? 
  // ah, maybe necessary for the intermediate results, should be private
  // MContainerKind _container;
  MAggregationKind _aggregation;
  MOrderingKind _ordering;
};

// ---------------------------------------------------------------

/** A C++ type is composed of TypeComponents.
 *
 * "const char *" is represented by Pointer(Const(Fundamental))
 * "map<string, Employee *>" by TemplateClass(Class, Pointer(Class))
 * But not all tokens have a separate component, eg.
 * "unsigned long int" is just Fundamental.
 *
 * These classes are vital for the separation of type parsing 
 * and type analysis. */
/** @dia:pos -159,55 */
class MTypeComponent {
public:
  /** Describes itself.
   * @param result Where results are accumulated
   * @param ns Used for class lookups
   */
  virtual void Analyze(MTypeAnalysis &result, MNamespace *ns) = 0;
  virtual ~MTypeComponent() {}
private:
};

/** A fundamental type, made of simple type specifiers */
/** @dia:pos -180,83 */
/** @dia:route MTypeComponent;v,-144.056,58.6,65,-149,81,-165.056,83 */
class MTCFundamental : public MTypeComponent {
public:
  MTCFundamental(const string &s) : _name(s) {}
  /** Appends eg. "int" to "unsigned long" 
   * @param s a simple type specifier (C++ 7.1.5.2) */
  void Append(const string &s) { _name += " " + s; }
  virtual void Analyze(MTypeAnalysis &result, MNamespace *ns);
private:
  /** The type name, such as "signed char" */
  string _name;
};

class MTCClass;

/** A namespace prefix, such as "Foo::Bar::" 
 *
 * Used for construction of qualified classes and member pointers */
/** @dia:pos -106,67 */
/** @dia:route MTypeComponent;v,-144.056,58.6,65,-92.9952,67 */
class MTCNameSpecifier : public MTypeComponent {
public:
  /* Constructs "aclass::" from "aclass" */
  MTCNameSpecifier(MTCClass *aclass);
  string GetName() { return _name; }
  virtual void Analyze(MTypeAnalysis &, MNamespace *) {} //dummy
private:
  string _name;
};


/** A target class */
/** @dia:pos -148,67 */
/** @dia:route MTypeComponent;v,-144.056,58.6,65,-133.056,67 */
class MTCClass : public MTypeComponent {
public:
  MTCClass(const string &name) : _name(name) {}
  MTCClass(const string &name, MTCNameSpecifier *q);
  string GetName() { return _name; }
  virtual void Analyze(MTypeAnalysis &result, MNamespace *ns);
protected:
  string _name;
};

/** A template class. Medoosa assumes it is a container */
/** @dia:pos -148,81 */
/** @dia:route MTCClass;v,-133.056,72.8,74,-133.056,81 */
class MTCTemplateClass : public MTCClass {
public:
  MTCTemplateClass(MTCClass *aclass) : MTCClass(*aclass) { delete aclass; }
  virtual ~MTCTemplateClass();
  void AddParam(MTypeComponent *param) { _parameters.push_back(param); }
  virtual void Analyze(MTypeAnalysis &result, MNamespace *ns);
private:
  /** @dia:route 4,7;h,-118.112,81.7,-115,58.6,-129.112 */
  /** @uml:order ordered */
  list<MTypeComponent *> _parameters; // polymorphic
};

/** A reference, as in "string &s" */
/** @dia:pos -106,74 */
/** @dia:route MTypeComponent;v,-144.056,58.6,65,-108,73,-91.056,74 */
class MTCReference : public MTypeComponent {
public:
  MTCReference(MTypeComponent *target): _target(target) {}
  virtual ~MTCReference() { delete _target; }
  virtual void Analyze(MTypeAnalysis &result, MNamespace *ns);
private:
  /** @dia:route 8,4;h,-106,76.3,-113,55.7,-129.112 */
  MTypeComponent *_target;
};

/** A pointer */
/** @dia:pos -106,82 */
/** @dia:route MTypeComponent;v,-144.056,58.6,65,-108,81,-91.056,82 */
class MTCPointer : public MTypeComponent {
public:
  MTCPointer(MTypeComponent *target): _target(target) {}
  virtual ~MTCPointer() { delete _target; }
  virtual void Analyze(MTypeAnalysis &result, MNamespace *ns);
private:
  /** @dia:route 8,9;h,-106,84.3,-113,57.3,-129.112 */
  MTypeComponent *_target;
};


/** A pointer to a member of a class.
 *
 * Note that it is not a descendant of MTCPointer, 
 * because it does not imply an association.
 * @see MTCPointer */
/** @dia:pos -106,89 */
/** @dia:route MTypeComponent;v,-144.056,58.6,65,-108,88,-91.056,89 */
class MTCMemberPointer : public MTypeComponent {
public:
  virtual ~MTCMemberPointer() { delete _base; delete _target; }
  /** first phase of construction, due to parser design */
  MTCMemberPointer(MTCNameSpecifier *base): _target(NULL), _base(base) {}
  /** second phase of construction, due to parser design */
  void SetTarget(MTypeComponent *target) { assert(_target==NULL); _target = target; }
  virtual void Analyze(MTypeAnalysis &result, MNamespace *ns);
private:
  /** For "int Foo::* p", the target is int */
  /** @dia:route 8,2;h,-106,91.3,-113,55,-129.112 */
  MTypeComponent *_target;
  /** For "int Foo::* p", the base is Foo:: */
  /** @dia:route 9,9;h,-76.112,91.3,-73,68.9,-79.9904 */
  /** @uml:mul 1 */
  MTCNameSpecifier *_base;
};

/** An array */
/** @dia:pos -180,67 */
/** @dia:route MTypeComponent;v,-144.056,58.6,65,-165.056,67 */
class MTCArray : public MTypeComponent {
public:
  MTCArray(MTypeComponent *target, const string& bound);
  virtual ~MTCArray() { delete _target; }
  virtual void Analyze(MTypeAnalysis &result, MNamespace *ns);
private:  
  /** @dia:route 10,8;h,-180,70.1,-182,57.3,-159 */
  /** @uml:mul 1 */
  MTypeComponent *_target;
  /** @dia:route 5,0;h,-95,61,-181,67,-180 */
  MMultiplicity _multiplicity; // does not fit precisely
};

/** A pointer to a function. 
 *
 * The parameters of the function are not modeled. */
/** @dia:pos -180,75 */
/** @dia:route MTypeComponent;v,-144.056,58.6,65,-149,73,-165.056,75 */
class MTCFunctionPointer : public MTypeComponent {
public:
  MTCFunctionPointer(MTypeComponent *target): _return(target) {}
  virtual ~MTCFunctionPointer() { delete _return; }
  virtual void Analyze(MTypeAnalysis &result, MNamespace *ns);
private:
  /** @dia:route 3,3;h,-180,75.7,-183,55.7,-159 */
  /** @uml:mul 1 */
  MTypeComponent *_return;
  // parameter  types... but they are not interesting
};

// ---------------------------------------------------------------

#endif
